/*----------------------------------------------------------
   Puts mail into Weasel's "forward" directory, so that it
   will be sent out the next time Weasel is on line.
   (If Weasel is already on line, it will check this
   directory every few minutes for new items.)

           Author:       Peter Moylan (peter@pmoylan.org)
           Started:      13 May 2000
           Last revised: 16 November 2020

           Modified by:  Steven Levine (steve53@earthlink.net)

   2020-11-21 SHL Localize CheckPrerequisites
   2020-11-21 SHL Clarify usage notes
   2020-11-21 SHL Warn if reading from STDIN
   2022-03-09 SHL Correct CopyHeader continuation line handling
   2022-03-09 SHL Normalize quoting
   2022-03-09 SHL Normalize keyword case
   2022-03-09 SHL Support multiple, standard addresses in CleanupAddress
   2022-03-24 SHL Support more addresses formats in CleanupAddress
   2022-04-02 SHL Handle extra commas in CleanupAddress
   2022-09-30 SHL Update CopyHeader unfolding logic to ensure last header line parsed
   2022-12-01 SHL Check input file exists and can open and die if not

   Usage:
           qmail -f sender@localdomain -t recipient@remotedomain filename

           where the argument after the '-f' specifies the sender;
           the argument after the -t specifies the address to which
           the mail is addressed; and filename is the file that is
           to be sent.  (This file must contain both the header and
           the body of the message.)
           The arguments may be given in any order and are all
           optional.
           If the -f and/or -t arguments are missing then the
           addresses are taken from the From: and/or To: lines in
           the header of the message.  If the filename is missing
           then the message is taken from standard input.  If
           From: and To: header lines are missing, they are added
           using the -f and -t arguments if present.

           The current working directory must contain the file
           WEASEL.INI or WEASEL.TNI, because we need to
           work out where the "forward" directory is.

   Prerequisites:

       1.  The Rexx scripts SelectTNI.cmd and INI_get.cmd must
           must be either in the current directory or in your
           PATH.  To get these scripts, look for the latest version
           of GenINI.zip in the same place you found qmail.zip.

       2.  You must have RXU.DLL in your LIBPATH.  RXU, a Rexx
           library written by Dave Boll, is available from Hobbes
           with the name rxu1a.zip.  (But is possible that you
           already have it for use with another Rexx utility.)

       3.  The executable SayTZ.exe, which is in qmail.zip,
           must be placed either in the current directory or
           in your PATH.

------------------------------------------------------------*/

/****************************************************************/
/*                       MAIN PROGRAM                           */
/****************************************************************/

CALL RxFuncAdd SysLoadFuncs, 'REXXUTIL', SysLoadFuncs
CALL SysLoadFuncs
CALL CheckPrerequisites rxu SelectTNI INI_get saytz.exe

Globals = 'ReturnPath ForwardPath InFile'

PARSE ARG userargs
CALL ParseArgs userargs

Nul = '00'X
IF SelectTNI('Weasel') > 0 THEN INIname = 'Weasel.TNI'
ELSE INIname = 'Weasel.INI'
SAY 'Using 'INIname
MailRoot = INI_get(INIname, '$SYS', 'MailRoot')
j = POS(Nul,MailRoot)
IF j > 0 THEN MailRoot = LEFT(MailRoot, j-1)
MailRoot = TRANSLATE(MailRoot, '\', '/')
MailRoot = STRIP(MailRoot)

TmpFile = SysTempFileName(MailRoot||'forward\T?????.tmp')
IF TmpFile = '' THEN
    DO
        SAY 'Failed to find forward directory 'MailRoot||'forward'
        EXIT 1
    END

/* Copy the original input file to a new file, inserting '.'    */
/* fill if needed.  This also lets us deal with input from      */
/* the keyboard.                                                */

IF InFile = '' THEN DO
    InFile = 'STDIN'
    SAY 'Reading message from STDIN'
END
ELSE DO
    CALL STREAM InFile, 'C', 'QUERY EXISTS'
    IF RESULT = '' THEN  DO
        SAY 'Cannot access input file' INFILE
        EXIT 1
    END
    CALL STREAM InFile, 'C', 'OPEN READ'
    IF RESULT \= 'READY:' THEN  DO
        SAY 'Cannot open input file' INFILE ' for read'
        EXIT 1
    END
END

NewFile = SysTempFileName(MailRoot||'forward\N?????.tmp')
IF NewFile = '' THEN
    DO
        SAY "Can't create temporary file"
        '@DEL 'TmpFile
        EXIT 1
    END

/* The header requires special handling. */

CALL CopyHeader NewFile

/* Now copy the body from InFile to NewFile. */

DO FOREVER
    buffer = LINEIN(InFile)
    IF STREAM(InFile, 'S') \= 'READY' THEN LEAVE
    /* Double leading dot */
    IF LEFT(buffer, 1) = '.' THEN CALL CHAROUT NewFile, '.'
    CALL LINEOUT NewFile, buffer
END
CALL STREAM InFile, 'C', 'CLOSE'
CALL STREAM NewFile, 'C', 'CLOSE'

InFile = NewFile

    /********************************************************************/
    /* While relay mail is waiting to be sent it is stored as a file    */
    /* which starts with the following details.                         */
    /*   4 bytes   format version, value 'V000'                         */
    /*   4 bytes   send time                                            */
    /*   1 byte    retry number                                         */
    /*   1 byte    Boolean notify-on-failure flag                       */
    /*   variable  sender (character string)                            */
    /*   variable  recipient list, bounded by () and comma-separated    */
    /*                                                                  */
    /* The message content starts immediately after this.               */
    /********************************************************************/

Nul = D2C(0)
CALL CHAROUT TmpFile, 'V000'||Nul||Nul||Nul||Nul||Nul||'1'X
CALL CHAROUT TmpFile, ReturnPath||'('
CALL CHAROUT TmpFile, ForwardPath||')'
CALL STREAM TmpFile, 'C', 'CLOSE'
'@COPY 'TmpFile'/B+'InFile'/A 'TmpFile'/B >NUL'
'@RENAME 'TmpFile' 'SysTempFileName('?????.FWD')
'@DEL 'InFile

/* Tell Weasel to re-check for outgoing mail. */

CALL RxFuncAdd 'RxuInit','RXU','RxuInit'
CALL RxuInit
SemName = '\SEM32\WEASEL\FORWARDMAIL'
IF RxOpenEventSem(hev, SemName) \= 0 THEN
    rc = RxCreateEventSem( hev ,'Shared', SemName, 'Reset')
CALL RxPostEventSem hev
CALL RxResetEventSem hev
CALL RxCloseEventSem hev

EXIT 0

/****************************************************************/
/*              PARSING THE COMMAND-LINE ARGUMENTS              */
/****************************************************************/

ParseArgs: PROCEDURE expose (Globals)
   ReturnPath = ''
   ForwardPath = ''
   InFile = ''
   PARSE ARG head tail
   head = STRIP(head)
   DO WHILE head \= ''
       IF head = '-f' THEN
           DO
              PARSE VAR tail ReturnPath head tail
              ReturnPath = STRIP(ReturnPath)
           END
       ELSE IF head = '-t' THEN
           DO
              PARSE VAR tail ForwardPath head tail
              ForwardPath = STRIP(ForwardPath)
           END
       ELSE
           DO
              InFile = head
              PARSE VAR tail head tail
           END
       head = STRIP(head)
   END
   RETURN

/****************************************************************/
/*      COPYING THE HEADER, WITH SPECIAL HANDLING OF THE        */
/*               "To:" AND "From:" INFORMATION                  */
/****************************************************************/

CopyHeader: PROCEDURE EXPOSE (Globals)

    PARSE ARG NewFile
    HaveDate = 0
    HaveForwardPath = ForwardPath \= ''
    HaveReturnPath = ReturnPath \= ''
    HaveFROM = 0
    HaveTO = 0
    HaveNextLine = 0
    UnfoldedLine = ''

    /* Need to read ahead to properly process continuation lines
       We assume input file is not empty and last line is properly terminated
    */

    DO FOREVER

        /* Each time around this loop we look at one line of   */
        /* the header.                                         */

        IF HaveNextLine THEN DO
              line = NextLine
              HaveNextLine = 0
        END
        ELSE DO
            line = LINEIN(InFile)
            /* Delay output of end of header marker */
            IF line \== '' THEN
                CALL LINEOUT NewFile, line
        END

        /* 2022-03-09 SHL only blank line can end header */
        IF line == '' & UnfoldedLine == '' THEN
          LEAVE                         /* End of header or end of file */

        IF line \== '' & LEFT(line,1) == ' ' THEN DO
            /* Continuation line - accumulate */
            IF UnfoldedLine == '' THEN DO
                /* Something wrong with header - it's some one else's problem */
                ITERATE
            END
            /* First space is not part of original line */
            UnfoldedLine = UnfoldedLine || SUBSTR(line, 2)
            ITERATE
        END

        /* Start of new header line */
        IF UnfoldedLine == '' THEN DO
          UnfoldedLine = line
          ITERATE
        END

        NextLine = line
        HaveNextLine = 1
        line = UnfoldedLine
        UnfoldedLine = ''

        IF POS(':', line) = 0 THEN
            ITERATE                     /* Something wrong with header */

        PARSE VAR line keyword ':' remainder
        keyword = TRANSLATE(STRIP(keyword))

        IF keyword = 'DATE' THEN
            HaveDate = 1
        ELSE IF keyword = 'FROM' THEN DO
            HaveFROM = 1
            IF \HaveReturnPath THEN DO
                remainder = CleanupAddress(remainder)
                ReturnPath = remainder
                HaveReturnPath = 1
            END
        END
        ELSE IF keyword = 'TO' | keyword = 'CC' | keyword = 'BCC' THEN DO
            HaveTO = 1
            IF \HaveForwardPath THEN DO
                remainder = CleanupAddress(remainder)
                IF ForwardPath = '' THEN
                    ForwardPath = remainder
                ELSE
                    ForwardPath = ForwardPath || ',' || remainder
            END
        END
    END /* while in header */

    /* Add "From:" and "To:" header lines if they are missing. */

    IF \HaveFROM & HaveReturnPath THEN
        CALL LINEOUT NewFile, 'From: 'ReturnPath
    IF \HaveTO & HaveForwardPath THEN
        CALL LINEOUT NewFile, 'To: 'ForwardPath

    /* Add "Date:" if it is missing. */

    IF \HaveDate THEN DO
        /* RFC 2822 syntax is day, dd mm yyyy hh:mm:ss zzz  */
        /* The three-char day name is optional.             */
        /* The SayTZ program gives us the time zone         */

        '@SayTZ | RXQUEUE'
        PULL zzzz
        ts = LEFT(DATE('W'),3)', 'DATE()' 'TIME()' 'zzzz
        CALL LINEOUT NewFile, 'Date: 'ts
    END

    /* Put out the blank line that terminates the header, plus (in the  */
    /* abnormal case where the original file did not have a blank       */
    /* line) the line that we overshot by when reading the input.       */

    CALL LINEOUT NewFile, ''
    IF line \= '' THEN CALL LINEOUT NewFile, line

    RETURN

/****************************************************************/
/*                 TIDYING UP AN E-MAIL ADDRESS                 */
/****************************************************************/

CleanupAddress: PROCEDURE

    /* The argument to this procedure is an e-mail address from */
    /* the message header.  It has the form user@domain, plus   */
    /* possibly some other detail.  We return the user@domain,  */
    /* stripping off the rest.                                  */

    PARSE ARG line

    /* Strip quoted strings */
    DO FOREVER
        lquote = POS('"', line)
        IF lquote = 0 THEN LEAVE
        rquote = POS('"', line, lquote + 1)
        IF rquote = 0 THEN LEAVE        /* Missing trailing quote */
        line = LEFT(line, lquote - 1) || SUBSTR(line, rquote + 1)
    END

    /* Strip (...) strings */
    DO FOREVER
        lpar = POS('(', line)
        IF lpar = 0 THEN LEAVE
        rpar = POS(')', line, lpar + 1)
        IF rpar = lpar THEN LEAVE               /* Missing ) */
        line = LEFT(line, lpar - 1) || SUBSTR(line, rpar + 1)
    END

    more = 1
    comma = 0

    DO WHILE more

        more = 0

        /* Given comma separated address list with addresses of the form
             name <address>
           or
             address
           strip name and < and > and strip whitespace as needed.
         */
        nextComma = POS(',', line, comma + 1)
        IF nextComma = 0 THEN
            addr = SUBSTR(line, comma + 1)
        ELSE
            addr = SUBSTR(line, comma + 1, nextComma - comma - 1)
        lbar = POS('<', addr)
        IF lbar > 0 THEN DO
            rbar = POS('>', addr, lbar + 1)
            IF rbar > lbar THEN
                addr = SUBSTR(addr, lbar + 1, rbar - lbar - 1)
        END
        addr = STRIP(addr)
        /* Avoid empty addresses caused by multiple commas */
        IF nextComma = 0 THEN DO
            /* Append to end */
            IF addr == '' & comma > 0 THEN
              comma = comma - 1
            line = LEFT(line, comma) || addr
        END
        ELSE DO
            /* Insert in middle */
            IF addr == '' & comma > 0 THEN
              comma = comma - 1
            line = LEFT(line, comma) || addr || SUBSTR(line, nextComma)
            comma = comma + 1 + length(addr)
            more = 1                    /* Not last address */
        END

        if 0 & trace() <> '?A' then trace '?A' ; nop    /* FIXME to be gone debug */

    END /* more */

    RETURN line

/****************************************************************/
/*                      CHECKING PREREQUISITES                  */
/****************************************************************/

CheckPrerequisites: PROCEDURE

    /* The argument is a space-separated list of prerequisite   */
    /* functions, for example                                   */
    /*      CALL CheckPrerequisites rxu SelectTNI INIget        */
    /* where (at least in this version) each list item is       */
    /* either 'rxu' or a function from my TNItools package.     */
    /* If any is missing then we exit with an error message.    */

    PARSE UPPER ARG funclist
    funclist = STRIP(funclist)
    needrxu = 0
    needtools = 0
    needexe = 0
    DO WHILE funclist \= ''
        PARSE VAR funclist func funclist
        funclist = STRIP(funclist)
        IF func = 'RXU' THEN DO
            /* Initialise RXU if not already available, fail if */
            /* the RxFuncAdd operation fails.  We must          */
            /* RxFuncQuery RxuTerm because RxuTerm does not     */
            /* deregister RxuInit.  The RxFuncDrop is needed    */
            /* because RxFuncAdd seems to report failure if the */
            /* function is already registered.                  */

            IF RxFuncQuery('RxuTerm') THEN DO
                CALL RxFuncDrop('RxuInit')
                CALL RxFuncAdd 'RxuInit','RXU','RxuInit'
                IF result THEN DO
                    SAY 'Cannot load RXU'
                    needrxu = 1
                END
                ELSE CALL RxuInit
            END
        END
        ELSE DO
            /* Check script or executable already in PATH */
            ext = RIGHT(func, 4)
            IF ext \== '.CMD' & ext \== '.EXE' THEN DO
                ext = '.CMD'
                func = func || ext
            END
            s = SysSearchPath('PATH', func)
            IF s == '' THEN DO
                /* If script not in PATH and caller's directory not
                   in PATH and caller's directory not current directory,
                   add caller's directory to PATH and retry
                */
                PARSE SOURCE . . cmd
                iSlash = LASTPOS('\', cmd)
                IF iSlash < 2 THEN
                    dir = directory()   /* Should never occur */
                ELSE
                    dir = SUBSTR(cmd, 1, iSlash - 1)
                IF TRANSLATE(dir) \== TRANSLATE(directory()) THEN DO
                    env = 'OS2ENVIRONMENT'
                    path = VALUE('PATH',, env)
                    IF POS(TRANSLATE(dir) || ';', TRANSLATE(path) || ';') = 0 THEN DO
                        path = dir || ';' || path
                        path = VALUE('PATH', path, env)
                        SAY 'Adding' dir 'to PATH'
                        s = SysSearchPath('PATH', func)
                    END
                END
            END
            IF s == '' THEN DO
                SAY 'ERROR: 'func' must be in your PATH or in' dir
                IF ext == '.EXE' THEN
                  needexe = 1
                ELSE
                  needtools = 1
            END
        END
    END
    IF needrxu THEN SAY 'You can find RXU1a.zip at Hobbes'
    IF needtools THEN SAY 'Please install the TNItools package'
    IF needrxu | needexe | needtools THEN EXIT 1
    RETURN

/****************************************************************/

